program swe_letkf
    use shared
    implicit none

    double precision,DIMENSION(:,:,:,:),allocatable :: EnsembleU
    double precision,DIMENSION(:,:,:,:),allocatable :: EnsembleV
    double precision,DIMENSION(:,:,:,:),allocatable :: EnsemblePhi

    double precision,DIMENSION(:,:),allocatable :: data2DPhi
    double precision,DIMENSION(:,:),allocatable :: data2DU
    double precision,DIMENSION(:,:),allocatable :: data2DV

    ! X_p, X_f: N_{state} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: perturbationU
    double precision,DIMENSION(:,:),allocatable :: perturbationV
    double precision,DIMENSION(:,:),allocatable :: perturbationPhi

    ! Y_f, Y_p: N_{obs} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: ensObsU
    double precision,DIMENSION(:,:),allocatable :: ensObsV
    double precision,DIMENSION(:,:),allocatable :: ensObsPhi

    ! \bar{x}: N_{state}
    double precision,DIMENSION(:),allocatable :: meanU
    double precision,DIMENSION(:),allocatable :: meanV
    double precision,DIMENSION(:),allocatable :: meanPhi

    ! \bar{y}_f: N_{obs}
    double precision,DIMENSION(:),allocatable :: meanObsU
    double precision,DIMENSION(:),allocatable :: meanObsV
    double precision,DIMENSION(:),allocatable :: meanObsPhi

    ! X_p^\ell: N_{state}^\ell \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localPerturbationU
    double precision,DIMENSION(:,:),allocatable :: localPerturbationV
    double precision,DIMENSION(:,:),allocatable :: localPerturbationPhi

    ! \bar{x}_f^\ell: N_{state}^\ell
    double precision,DIMENSION(:),allocatable :: localMeanU
    double precision,DIMENSION(:),allocatable :: localMeanV
    double precision,DIMENSION(:),allocatable :: localMeanPhi

    ! Y_p^\ell: N_{obs}^\ell \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localEnsObsU
    double precision,DIMENSION(:,:),allocatable :: localEnsObsV
    double precision,DIMENSION(:,:),allocatable :: localEnsObsPhi

    ! \bar{y}_f^\ell: N_{obs}^\ell
    double precision,DIMENSION(:),allocatable :: localMeanObsU
    double precision,DIMENSION(:),allocatable :: localMeanObsV
    double precision,DIMENSION(:),allocatable :: localMeanObsPhi

    ! C: N_{ens} \times N_{obs}^\ell
    double precision,DIMENSION(:,:),allocatable :: localTempEnsObsU
    double precision,DIMENSION(:,:),allocatable :: localTempEnsObsV
    double precision,DIMENSION(:,:),allocatable :: localTempEnsObsPhi

    ! \tilde{P}_a: N_{ens} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localAnalysisCovarU
    double precision,DIMENSION(:,:),allocatable :: localAnalysisCovarV
    double precision,DIMENSION(:,:),allocatable :: localAnalysisCovarPhi

    ! D: N_{ens} -- eigenvalues of \tilde{P}_a = L D L^T
    double precision,DIMENSION(:),allocatable :: localEigvU
    double precision,DIMENSION(:),allocatable :: localEigvV
    double precision,DIMENSION(:),allocatable :: localEigvPhi

    ! work: N_{work} -- work matrix for lapack dsyev
    double precision,DIMENSION(:),allocatable :: localWork

    ! W_a: N_{ens} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEnsU
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEnsV
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEnsPhi

    ! intermediate var: N_{ens} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEns2U
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEns2V
    double precision,DIMENSION(:,:),allocatable :: localTempEnsEns2Phi

    ! \bar{w}: N_{ens}
    double precision,DIMENSION(:),allocatable :: localMeanTempEnsEnsU
    double precision,DIMENSION(:),allocatable :: localMeanTempEnsEnsV
    double precision,DIMENSION(:),allocatable :: localMeanTempEnsEnsPhi

    ! X_a^\ell: N_{state}^\ell \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: localAnalysisU
    double precision,DIMENSION(:,:),allocatable :: localAnalysisV
    double precision,DIMENSION(:,:),allocatable :: localAnalysisPhi

    ! local variables
    integer :: i, j, k, kpass, ierr, member, stateindex, NXCorrLength, NYCorrLength, maxls
    integer :: locali, localj, localn, member2, lwork, analysisPoint, ln
    double precision :: dist, uval, vval, phival

    character(len=64) :: rmseFileName, file_num

    logical analysisDone

    call initialize()

    ! Calculate maximum local size
    NXCorrLength = ceiling(CovLocalizationLength/DX)
    NYCorrLength = ceiling(CovLocalizationLength/DY)
    maxls = 2*NXCorrLength*2*NYCorrLength

    ! X: N_{ens} \times n \times m \times nt
    call alloc4(EnsembleU,NoMembers,n,m,nt,'EnsembleU')
    call alloc4(EnsembleV,NoMembers,n,m,nt,'EnsembleV')
    call alloc4(EnsemblePhi,NoMembers,n,m,nt,'EnsemblePhi')

    ! helper data: n-2 \times m-2
    call alloc2(data2DU,n-2,m-2,'data2DU')
    call alloc2(data2DV,n-2,m-2,'data2DV')
    call alloc2(data2DPhi,n-2,m-2,'data2DPhi')

    ! X_p, X_f: N_{state} \times N_{ens}
    call alloc2(perturbationU,ns,NoMembers,'perturbationU')
    call alloc2(perturbationV,ns,NoMembers,'perturbationV')
    call alloc2(perturbationPhi,ns,NoMembers,'perturbationPhi')

    ! Y_f, Y_p: N_{obs} \times N_{ens}
    call alloc2(ensObsU,nobs,NoMembers,'ensObsU')
    call alloc2(ensObsV,nobs,NoMembers,'ensObsV')
    call alloc2(ensObsPhi,nobs,NoMembers,'ensObsPhi')

    ! \bar{x}: N_{state}
    call alloc1(meanU,ns,'meanU')
    call alloc1(meanV,ns,'meanV')
    call alloc1(meanPhi,ns,'meanPhi')

    ! \bar{y}_f: N_{obs}
    call alloc1(meanObsU,nobs,'meanObsU')
    call alloc1(meanObsV,nobs,'meanObsV')
    call alloc1(meanObsPhi,nobs,'meanObsPhi')

    ! \tilde{P}_a: N_{ens} \times N_{ens}
    call alloc2(localAnalysisCovarU,NoMembers,NoMembers,'localAnalysisCovarU')
    call alloc2(localAnalysisCovarV,NoMembers,NoMembers,'localAnalysisCovarV')
    call alloc2(localAnalysisCovarPhi,NoMembers,NoMembers,'localAnalysisCovarPhi')

    ! D: N_{ens}
    call alloc1(localEigvU,NoMembers,'localEigvU')
    call alloc1(localEigvV,NoMembers,'localEigvV')
    call alloc1(localEigvPhi,NoMembers,'localEigvPhi')

    ! W_a: N_{ens} \times N_{ens}
    call alloc2(localTempEnsEnsU,NoMembers,NoMembers,'localTempEnsEnsU')
    call alloc2(localTempEnsEnsV,NoMembers,NoMembers,'localTempEnsEnsV')
    call alloc2(localTempEnsEnsPhi,NoMembers,NoMembers,'localTempEnsEnsPhi')

    ! temp: N_{ens} \times N_{ens}
    call alloc2(localTempEnsEns2U,NoMembers,NoMembers,'localTempEnsEns2U')
    call alloc2(localTempEnsEns2V,NoMembers,NoMembers,'localTempEnsEns2V')
    call alloc2(localTempEnsEns2Phi,NoMembers,NoMembers,'localTempEnsEns2Phi')

    ! \bar{w}: N_{ens}
    call alloc1(localMeanTempEnsEnsU,NoMembers,'localMeanTempEnsEnsU')
    call alloc1(localMeanTempEnsEnsV,NoMembers,'localMeanTempEnsEnsV')
    call alloc1(localMeanTempEnsEnsPhi,NoMembers,'localMeanTempEnsEnsPhi')

    call readFiles()

    do member=1,NoMembers

        print *, 'Reading the perturbations for member', member

        write(file_num,'(I0.4)') member

        open(unit=12,file='SWM_out_' // file_num,status='unknown', &
             form='formatted')
        read(12,*) ((data2DU(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DV(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DPhi(i,j),i=1,n-2),j=1,m-2)
        close(12)

        EnsembleU(member, 2:n-1, 2:m-1, 1) = data2DU(1:n-2, 1:m-2)
        EnsembleV(member, 2:n-1, 2:m-1, 1) = data2DV(1:n-2, 1:m-2)
        EnsemblePhi(member, 2:n-1, 2:m-1, 1) = data2DPhi(1:n-2, 1:m-2)
    end do

    ! calculate the mean of the ensemble at the initial time
    meanU = 0.
    meanV = 0.
    meanPhi = 0.

    ! collect values for the mean
    do member=1,NoMembers
	    do i=2,n-1
	        do j=2,m-1
	            stateindex = map_2to1(i-1,j-1,n-2)

	            meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, 1)
	            meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, 1)
	            meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, 1)
	        enddo
	    enddo
	enddo

	meanU = meanU / NoMembers
	meanV = meanV / NoMembers
	meanPhi = meanPhi / NoMembers

    do i=2,n-1
        do j=2,m-1
            stateindex = map_2to1(i-1,j-1,n-2)

            u(i,j,1) = meanU(stateindex)
            v(i,j,1) = meanV(stateindex)
            phi(i,j,1) = meanPhi(stateindex)
        enddo
    enddo

    do k=1,nt
        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        call swe_model(u(1:n, 1:m, kpass:k), &
                       v(1:n, 1:m, kpass:k), &
                       phi(1:n, 1:m, kpass:k), &
                       n, m, nt, kpass, k, k)

        call calculateRmse(k)
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

81  format('RMSE-LETKF-Obsop',i1,'CovL',F5.0,'txt')

    write(rmseFileName,81) ObsOpNum, CovLocalizationLength

    open(unit=84, file=rmseFileName, status='unknown', form='formatted')

    do k=1,nt
        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        meanU = 0.
        meanV = 0.
        meanPhi = 0.

        meanObsU = 0.
        meanObsV = 0.
        meanObsPhi = 0.

        do member=1,NoMembers
            ! advance the model forward in time
	        call swe_model(EnsembleU(member, 1:n, 1:m, kpass:k), &
	                       EnsembleV(member, 1:n, 1:m, kpass:k), &
	                       EnsemblePhi(member, 1:n, 1:m, kpass:k), &
	                       n, m, nt, kpass, k, k)

            ! collect values for the mean
            do i=2,n-1
                do j=2,m-1

		            stateindex = map_2to1(i-1,j-1,n-2)

		            meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)

                    call obs_operator(EnsembleU(member, i, j, k), &
                                      EnsembleV(member, i, j, k), &
                                      EnsemblePhi(member, i, j, k), &
                                      uval,vval,phival)

                    ensObsU(stateindex,member) = uval
                    ensObsV(stateindex,member) = vval
                    ensObsPhi(stateindex,member) = phival

                    meanObsU(stateindex) = meanObsU(stateindex) + uval
                    meanObsV(stateindex) = meanObsV(stateindex) + vval
                    meanObsPhi(stateindex) = meanObsPhi(stateindex) + phival
                enddo
            enddo
	    enddo

	    meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        meanObsU = meanObsU / NoMembers
        meanObsV = meanObsV / NoMembers
        meanObsPhi = meanObsPhi / NoMembers

	    do member=1,NoMembers
            do i=2,n-1
	            do j=2,m-1
		            stateindex = map_2to1(i-1,j-1,n-2)

	                perturbationU(stateindex,member) = EnsembleU(member, i, j, k) - &
	                    meanU(stateindex)

		            perturbationV(stateindex,member) = EnsembleV(member, i, j, k) - &
                        meanV(stateindex)

		            perturbationPhi(stateindex,member) = EnsemblePhi(member, i, j, k)- &
                        meanPhi(stateindex)

                    ensObsU(stateindex,member) = ensObsU(stateindex,member) - &
                        meanObsU(stateindex)

                    ensObsV(stateindex,member) = ensObsV(stateindex,member) - &
                        meanObsV(stateindex)

                    ensObsPhi(stateindex,member) = ensObsPhi(stateindex,member) - &
                        meanObsPhi(stateindex)
	            enddo
            enddo
	    enddo

        analysisDone = .false.

	    ! main loop over all grid points
        do i=2,n-1

            if (analysisDone) then
                exit
            endif

            do j=2,m-1
                if (analysisDone) then
                    exit
                endif

	            localn = 0

	            ! perform local analysis for each grid point
	            do locali = max(2,i-NXCorrLength),min(n-1,i+NXCorrLength)
                    do localj = max(2,j-NYCorrLength),min(m-1,j+NYCorrLength)

                        dist = sqrt(abs((DX*(locali-i))**2+(DY*(localj-j))**2))

                        if (dist .le. CovLocalizationLength) then
                            localn = localn + 1

                            if (locali .eq. i .and. localj .eq. j) then
                                analysisPoint = localn
                            endif
                        endif
                    enddo
                enddo

			    ! X_p^\ell: N_{state}^\ell \times N_{ens}
			    call alloc2(localPerturbationU,localn,NoMembers,'localPerturbationU')
                call alloc2(localPerturbationV,localn,NoMembers,'localPerturbationV')
                call alloc2(localPerturbationPhi,localn,NoMembers,'localPerturbationPhi')

			    ! \bar{x}_f^\ell: N_{state}^\ell
                call alloc1(localMeanU,localn,'localMeanU')
                call alloc1(localMeanV,localn,'localMeanV')
                call alloc1(localMeanPhi,localn,'localMeanPhi')

			    ! Y_p^\ell: N_{obs}^\ell \times N_{ens}
                call alloc2(localEnsObsU,localn,NoMembers,'localEnsObsU')
                call alloc2(localEnsObsV,localn,NoMembers,'localEnsObsV')
                call alloc2(localEnsObsPhi,localn,NoMembers,'localEnsObsPhi')

			    ! \bar{y}_f^\ell: N_{obs}^\ell
                call alloc1(localMeanObsU,localn,'localMeanObsU')
                call alloc1(localMeanObsV,localn,'localMeanObsV')
                call alloc1(localMeanObsPhi,localn,'localMeanObsPhi')

			    ! C: N_{ens} \times N_{obs}^\ell
			    call alloc2(localTempEnsObsU,NoMembers,localn,'localTempEnsObsU')
                call alloc2(localTempEnsObsV,NoMembers,localn,'localTempEnsObsV')
                call alloc2(localTempEnsObsPhi,NoMembers,localn,'localTempEnsObsPhi')

			    ! X_a^\ell: N_{state}^\ell \times N_{ens}
			    call alloc2(localAnalysisU,localn,NoMembers,'localAnalysisU')
                call alloc2(localAnalysisV,localn,NoMembers,'localAnalysisV')
                call alloc2(localAnalysisPhi,localn,NoMembers,'localAnalysisPhi')

			    localn = 0

                do locali = max(2,i-NXCorrLength),min(n-1,i+NXCorrLength)
                    do localj = max(2,j-NYCorrLength),min(m-1,j+NYCorrLength)

                        dist=sqrt(abs((DX*(locali-i))**2+(DY*(localj-j))**2))

                        if (dist .le. CovLocalizationLength) then
                            localn = localn + 1

                            stateindex = map_2to1(locali-1,localj-1,n-2)

                            localPerturbationU(localn,:) = perturbationU(stateindex,:)
                            localPerturbationV(localn,:) = perturbationV(stateindex,:)
                            localPerturbationPhi(localn,:) = perturbationPhi(stateindex,:)

                            localMeanU(localn) = meanU(stateindex)
                            localMeanV(localn) = meanV(stateindex)
                            localMeanPhi(localn) = meanPhi(stateindex)

                            localEnsObsU(localn,:) = ensObsU(stateindex,:)
                            localEnsObsV(localn,:) = ensObsV(stateindex,:)
                            localEnsObsPhi(localn,:) = ensObsPhi(stateindex,:)

                            localMeanObsU(localn) = uobs(locali,localj,k) - meanObsU(stateindex)
                            localMeanObsV(localn) = vobs(locali,localj,k) - meanObsV(stateindex)
                            localMeanObsPhi(localn) = phiobs(locali,localj,k) - meanObsPhi(stateindex)
                        endif
                    enddo
                enddo

                localAnalysisCovarU = 0.
                localAnalysisCovarV = 0.
                localAnalysisCovarPhi = 0.

                ! \tilde{P}_a = \left[ \frac{N_{ens}}{\lambda}I + \left(Y_p^\ell\right)^\trans R^{-1} Y_p^\ell\right]^{-1}
                do member=1,NoMembers
                    localAnalysisCovarU(member,member) = NoMembers/CovInflation
                    localAnalysisCovarV(member,member) = NoMembers/CovInflation
                    localAnalysisCovarPhi(member,member) = NoMembers/CovInflation
                enddo

                localAnalysisCovarU = localAnalysisCovarU + &
                    matmul(transpose(localEnsObsU)/(UObsStddev**2),localEnsObsU)

                localAnalysisCovarV = localAnalysisCovarV + &
                    matmul(transpose(localEnsObsV)/(VObsStddev**2),localEnsObsV)

                localAnalysisCovarPhi = localAnalysisCovarPhi + &
                    matmul(transpose(localEnsObsPhi)/(PhiObsStddev**2),localEnsObsPhi)

                call eigdecomp(NoMembers,localAnalysisCovarU,localEigvU)
                call eigdecomp(NoMembers,localAnalysisCovarV,localEigvV)
                call eigdecomp(NoMembers,localAnalysisCovarPhi,localEigvPhi)

                ! inverse of \tilde{P}_a^\ell is just the inverse eigenvalues
                localEigvU = 1./localEigvU
                localEigvV = 1./localEigvV
                localEigvPhi = 1./localEigvPhi

                localTempEnsEnsU = 0.
                localTempEnsEnsV = 0.
                localTempEnsEnsPhi = 0.

                ! get diagonal terms for W_a = sqrt((NoMembers-1)*localEigv) and \tilde{P}_a
                do member=1,NoMembers
                    localTempEnsEnsU(member,member) = sqrt((NoMembers - 1)*localEigvU(member))
                    localTempEnsEnsV(member,member) = sqrt((NoMembers - 1)*localEigvV(member))
                    localTempEnsEnsPhi(member,member) = sqrt((NoMembers - 1)*localEigvPhi(member))

                    localTempEnsEns2U(member,member) = localEigvU(member)
                    localTempEnsEns2V(member,member) = localEigvV(member)
                    localTempEnsEns2Phi(member,member) = localEigvPhi(member)
                enddo

                ! complete W_a = LD^{1/2}L^T
                localTempEnsEnsU = matmul(localAnalysisCovarU, &
                    matmul(localTempEnsEnsU,transpose(localAnalysisCovarU)))

                localTempEnsEnsV = matmul(localAnalysisCovarV, &
                    matmul(localTempEnsEnsV,transpose(localAnalysisCovarV)))

                localTempEnsEnsPhi = matmul(localAnalysisCovarPhi, &
                    matmul(localTempEnsEnsPhi,transpose(localAnalysisCovarPhi)))

                ! fill out \tilde{P}_a: multiply by LL^T
                localAnalysisCovarU = matmul(localAnalysisCovarU, &
                    matmul(localTempEnsEns2U,transpose(localAnalysisCovarU)))

                localAnalysisCovarV = matmul(localAnalysisCovarV, &
                    matmul(localTempEnsEns2V,transpose(localAnalysisCovarV)))

                localAnalysisCovarPhi = matmul(localAnalysisCovarPhi, &
                    matmul(localTempEnsEns2Phi,transpose(localAnalysisCovarPhi)))

                ! \bar{w}_a = \tilde{P}_a \left(Y_p^\ell\right)^\trans R^{-1} \bar{y}_f^\ell
                localMeanTempEnsEnsU = matmul(localAnalysisCovarU, &
                    matmul(transpose(localEnsObsU)/(UObsStddev**2),localMeanObsU))

                localMeanTempEnsEnsV = matmul(localAnalysisCovarV, &
                    matmul(transpose(localEnsObsV)/(VObsStddev**2),localMeanObsV))

                localMeanTempEnsEnsPhi = matmul(localAnalysisCovarPhi, &
                    matmul(transpose(localEnsObsPhi)/(PhiObsStddev**2),localMeanObsPhi))

                ! (W_a)_i = \left(W_a\right)_i + \bar{w}_a
                do member=1,NoMembers
                    localTempEnsEnsU(:,member) = localTempEnsEnsU(:,member) + &
                        localMeanTempEnsEnsU(:)

                    localTempEnsEnsV(:,member) = localTempEnsEnsV(:,member) + &
                        localMeanTempEnsEnsV(:)

                    localTempEnsEnsPhi(:,member) = localTempEnsEnsPhi(:,member) + &
                        localMeanTempEnsEnsPhi(:)
                enddo

                localAnalysisU = matmul(localPerturbationU, localTempEnsEnsU)
                localAnalysisV = matmul(localPerturbationV, localTempEnsEnsV)
                localAnalysisPhi = matmul(localPerturbationPhi, localTempEnsEnsPhi)

                do member=1,NoMembers
                    localAnalysisU(:,member) = localAnalysisU(:,member) + &
                        localMeanU(:)

                    localAnalysisV(:,member) = localAnalysisV(:,member) + &
                        localMeanV(:)

                    localAnalysisPhi(:,member) = localAnalysisPhi(:,member) + &
                        localMeanPhi(:)
                enddo

                if (NXCorrLength .ge. n .and. NYCorrLength .ge. m) then
                    ln = 0

                    do locali = 2,n-1
                        do localj = 2,m-1
                            ln = ln + 1

			                EnsembleU(:, locali, localj, k) = &
                                localAnalysisU(ln,:)

			                EnsembleV(:, locali, localj, k) = &
			                    localAnalysisV(ln,:)

			                EnsemblePhi(:, locali, localj, k) = &
			                     localAnalysisPhi(ln,:)

                            analysisDone = .true.
                        enddo
                    enddo
                else
	                EnsembleU(:, i, j, k) = localAnalysisU(analysisPoint,:)

	                EnsembleV(:, i, j, k) = localAnalysisV(analysisPoint,:)

	                EnsemblePhi(:, i, j, k) = localAnalysisPhi(analysisPoint,:)
                endif

				deallocate(localPerturbationU)
				deallocate(localPerturbationV)
				deallocate(localPerturbationPhi)

				deallocate(localMeanU)
				deallocate(localMeanV)
				deallocate(localMeanPhi)

				deallocate(localEnsObsU)
				deallocate(localEnsObsV)
				deallocate(localEnsObsPhi)

				deallocate(localTempEnsObsU)
                deallocate(localTempEnsObsV)
                deallocate(localTempEnsObsPhi)

				deallocate(localMeanObsU)
				deallocate(localMeanObsV)
				deallocate(localMeanObsPhi)

				deallocate(localAnalysisU)
				deallocate(localAnalysisV)
				deallocate(localAnalysisPhi)

                ! end loop over grid points
            enddo
	    enddo

	    meanU = 0.
        meanV = 0.
        meanPhi = 0.

        do member=1,NoMembers
             call set_boundary_conditions(k, m, n, EnsembleU(  member, :, :, k), &
                                                   EnsembleV(  member, :, :, k), &
                                                   EnsemblePhi(member, :, :, k))

            do i=2,n-1
                do j=2,m-1
                    stateindex = map_2to1(i-1,j-1,n-2)

                    meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)
                enddo
            enddo
        enddo

        meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        do i=2,n-1
	        do j=2,m-1
                stateindex = map_2to1(i-1,j-1,n-2)

	            u(i,j,k) = meanU(stateindex)
                v(i,j,k) = meanV(stateindex)
                phi(i,j,k) = meanPhi(stateindex)
            enddo
        enddo

        call calculateRmse(k)

        write (84,*) k, sqrt(urmse)/rmseroot, sqrt(vrmse)/rmseroot, sqrt(phirmse)/rmseroot

        print *, 'Completed iteration', k
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

    print *, 'Final RMSE is: '

    do k=1,nt
        call calculateRmse(k)
    enddo

    close(84)

    call finalize()

    deallocate(perturbationU)
	deallocate(perturbationV)
	deallocate(perturbationPhi)

	deallocate(ensObsU)
	deallocate(ensObsV)
	deallocate(ensObsPhi)

	deallocate(meanU)
	deallocate(meanV)
	deallocate(meanPhi)

	deallocate(meanObsU)
	deallocate(meanObsV)
	deallocate(meanObsPhi)

	deallocate(localAnalysisCovarU)
	deallocate(localAnalysisCovarV)
	deallocate(localAnalysisCovarPhi)

	deallocate(localTempEnsEnsU)
	deallocate(localTempEnsEnsV)
	deallocate(localTempEnsEnsPhi)

    deallocate(localTempEnsEns2U)
    deallocate(localTempEnsEns2V)
    deallocate(localTempEnsEns2Phi)

	deallocate(localMeanTempEnsEnsU)
	deallocate(localMeanTempEnsEnsV)
	deallocate(localMeanTempEnsEnsPhi)

end program

subroutine eigdecomp(N,A,eigv)
    implicit none

    integer, intent(in) :: N
    double precision, intent(inout), dimension(N,N) :: A
    double precision, intent(out), dimension(N) :: eigv

    integer :: lwork, ierr
    double precision, dimension(1) :: temp
    double precision, dimension(:), allocatable :: work

    ! query the optimal work space
    lwork = -1
    call DSYEV('V','U',N,A,N,eigv,temp,lwork,ierr)

    lwork = temp(1)

    allocate(work(lwork),stat=ierr)
    if (ierr.ne.0) then
        write(*,*) "Allocation error: work"
        STOP
    end if

    ! solve the eigenvalue problem
    call DSYEV('V','U',N,A,N,eigv,work,lwork,ierr)
    if (ierr.ne.0) then
        write(*,*) "Error solving the eigenvalue decomposition: ",ierr
        STOP
    end if

    deallocate(work)

end subroutine
