program swe_mlef
    use shared
    use mlef_mod
    implicit none

! L-BFGS variables
    integer :: ntrav, msave, nxm, nwork

    common /va15dd/ mp, lp, gtol

    integer :: mp, lp
    double precision :: gtol, oldcost

    double precision, allocatable, dimension(:) :: diag, ss, yy, ww
    double precision :: dzs(10), apd

    integer :: iprint(2), point
    logical :: diagco

! LMBM variables
    integer :: na, mcu, nw
    integer :: ipar(7),iout(3)

    double precision, allocatable, dimension(:) :: w
    double precision :: rpar(8)
    double precision :: time,rtim(2)
    double precision :: eps
    double precision, parameter :: MIN_EIGV = 1.d-6
    integer :: mc, mcinit

! DESCON variables
    integer nexp, iter, irs, fgcnt, lscnt, ibeta, itheta
    integer itert, irstot, fgtot,lstot, stoptest
    double precision gnorm, epsm
    double precision dc, cc

    logical angle, powell

! CG-Descent variables
      double precision, allocatable, dimension(:) :: dtemp, xtemp, gtemp
      integer :: fcnt, gcnt
      external :: funder_costonly, funder_gradonly

! local variables
    double precision, allocatable, dimension(:) :: grad, ksi, ksiv

    integer :: i, j, k, kpass, ierr, member, stateindex
    integer :: locali, localj, member2, lwork, ln
    double precision :: dist, uval, vval, phival
    integer :: icall, iflag, ifunder
    double precision :: costval,costu,costv,costp

    character(len=64) :: costFileName, rmseFileName
    character(len=64) :: file_num

    logical analysisDone

    call initialize()
    call mlef_initialize()

    call readFiles()

    nctr = NoMembers

! initialize L-BFGS variables
    ntrav = 25*nctr
    msave = 9
    nxm = msave*nctr
    nwork = nctr+2*msave

    allocate (diag(nctr), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for diag"

    allocate (ss(nxm), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for ss"

    allocate (yy(nxm), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for yy"

    allocate (ww(nwork), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for ww"

! initialize LMBM variables
    na = 4
    mcu = 9
    nw = 1 + 9*nctr + 2*nctr*na + 3*na + 2*nctr*(mcu+1) + 3*(mcu+2)*(mcu+1)/2 + 9*(mcu+1)

    allocate (w(nw), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for w"

! initialize CG-descent variables
      allocate (dtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for dtemp"

      allocate (xtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for xtemp"

      allocate (gtemp(nctr), STAT=ierr)
      if(ierr /= 0) write(*,*) "allocation error for gtemp"

! initialize local variables
    allocate (grad(nctr), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for grad"

    allocate (ksi(3*nctr), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for ksi"

    allocate (ksiv(nctr), STAT=ierr)
    if(ierr /= 0) write(*,*) "allocation error for ksiv"

    do member=1,NoMembers

        print *, 'Reading the perturbations for member', member

        write(file_num,'(I0.4)') member

        open(unit=12,file='SWM_out_' // file_num,status='unknown', &
             form='formatted')
        read(12,*) ((data2DU(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DV(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DPhi(i,j),i=1,n-2),j=1,m-2)
        close(12)

        EnsembleU(member, 2:n-1, 2:m-1, 1) = data2DU(1:n-2, 1:m-2)
        EnsembleV(member, 2:n-1, 2:m-1, 1) = data2DV(1:n-2, 1:m-2)
        EnsemblePhi(member, 2:n-1, 2:m-1, 1) = data2DPhi(1:n-2, 1:m-2)
    end do

    ! calculate the mean of the ensemble at the initial time
    meanU = 0.
    meanV = 0.
    meanPhi = 0.

    ! collect values for the mean
    do member=1,NoMembers
        do i=2,n-1
            do j=2,m-1
                stateindex = map_2to1(i-1,j-1,n-2)

                meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, 1)
                meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, 1)
                meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, 1)
            enddo
        enddo
    enddo

    meanU = meanU / NoMembers
    meanV = meanV / NoMembers
    meanPhi = meanPhi / NoMembers

    do i=2,n-1
        do j=2,m-1
            stateindex = map_2to1(i-1,j-1,n-2)

            u(i,j,1) = meanU(stateindex)
            v(i,j,1) = meanV(stateindex)
            phi(i,j,1) = meanPhi(stateindex)
        enddo
    enddo

    do k=1,nt
        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        call swe_model(u(1:n, 1:m, kpass:k), &
                       v(1:n, 1:m, kpass:k), &
                       phi(1:n, 1:m, kpass:k), &
                       n, m, nt, kpass, k, k)

        call calculateRmse(k)
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

 80 format('Cost-MLEF-',a4,'-Obsop',i1,'-Smoothing',i1,'.txt')
 81 format('RMSE-MLEF-',a4,'-Obsop',i1,'-Smoothing',i1,'CovL',F5.0,'txt')

    if (OptimizerNum .eq. 1) then ! L-BFGS

        write(costFileName,80) 'LMQN', ObsOpNum, ObsOpSmoothing
        write(rmseFileName,81) 'LMQN', ObsOpNum, ObsOpSmoothing, CovLocalizationLength

    elseif (OptimizerNum .eq. 2) then ! LMBM

        write(costFileName,80) 'LMBM', ObsOpNum, ObsOpSmoothing
        write(rmseFileName,81) 'LMBM', ObsOpNum, ObsOpSmoothing, CovLocalizationLength

    elseif (OptimizerNum .eq. 3) then ! DESCON

        write(costFileName,80) 'CGDC', ObsOpNum, ObsOpSmoothing
        write(rmseFileName,81) 'CGDC', ObsOpNum, ObsOpSmoothing, CovLocalizationLength

    elseif (OptimizerNum .eq. 4) then ! CG-Descent

        write(costFileName,80) 'CGDT', ObsOpNum, ObsOpSmoothing
        write(rmseFileName,81) 'CGDT', ObsOpNum, ObsOpSmoothing, CovLocalizationLength

    else
        print *,'Unknown OptimizerNum',OptimizerNum
        stop
    endif

    open(unit=84, file=costFileName, status='unknown', form='formatted')
    open(unit=86, file=rmseFileName, status='unknown', form='formatted')

    do k=1,nt
        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        meanU = 0.
        meanV = 0.
        meanPhi = 0.

        meanObsU = 0.
        meanObsV = 0.
        meanObsPhi = 0.

        do member=1,NoMembers
            ! advance the model forward in time
            call swe_model(EnsembleU(member, 1:n, 1:m, kpass:k), &
                           EnsembleV(member, 1:n, 1:m, kpass:k), &
                           EnsemblePhi(member, 1:n, 1:m, kpass:k), &
                           n, m, nt, kpass, k, k)

            ! collect values for the mean
            do i=2,n-1
                do j=2,m-1

                    stateindex = map_2to1(i-1,j-1,n-2)

                    meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)

                    call obs_operator(EnsembleU(member, i, j, k), &
                                      EnsembleV(member, i, j, k), &
                                      EnsemblePhi(member, i, j, k), &
                                      uval,vval,phival)

                    ensObsU(stateindex,member) = uval
                    ensObsV(stateindex,member) = vval
                    ensObsPhi(stateindex,member) = phival

                    meanObsU(stateindex) = meanObsU(stateindex) + uval
                    meanObsV(stateindex) = meanObsV(stateindex) + vval
                    meanObsPhi(stateindex) = meanObsPhi(stateindex) + phival
                enddo
            enddo
        enddo

        meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        meanObsU = meanObsU / NoMembers
        meanObsV = meanObsV / NoMembers
        meanObsPhi = meanObsPhi / NoMembers

        do member=1,NoMembers
            do i=2,n-1
                do j=2,m-1
                    stateindex = map_2to1(i-1,j-1,n-2)

                    perturbationU(stateindex,member) = EnsembleU(member, i, j, k) - &
                        meanU(stateindex)

                    perturbationV(stateindex,member) = EnsembleV(member, i, j, k) - &
                        meanV(stateindex)

                    perturbationPhi(stateindex,member) = EnsemblePhi(member, i, j, k)- &
                        meanPhi(stateindex)

                    ensObsU(stateindex,member) = ensObsU(stateindex,member) - &
                        meanObsU(stateindex)

                    ensObsV(stateindex,member) = ensObsV(stateindex,member) - &
                        meanObsV(stateindex)

                    ensObsPhi(stateindex,member) = ensObsPhi(stateindex,member) - &
                        meanObsPhi(stateindex)
                enddo
            enddo
        enddo

        analysisDone = .false.

        ! main loop over all grid points
        do i=2,n-1

            if (analysisDone) then
                exit
            endif

            do j=2,m-1
                if (analysisDone) then
                    exit
                endif

                localn = 0

                ! perform local analysis for each grid point
                do locali = max(2,i-NXCorrLength),min(n-1,i+NXCorrLength)
                    do localj = max(2,j-NYCorrLength),min(m-1,j+NYCorrLength)

                        dist = sqrt(abs((DX*(locali-i))**2+(DY*(localj-j))**2))

                        if (dist .le. CovLocalizationLength) then
                            localn = localn + 1

                            if (locali .eq. i .and. localj .eq. j) then
                                analysisPoint = localn
                            endif
                        endif
                    enddo
                enddo

                ! X_p^\ell: N_{state}^\ell \times N_{ens}
                call alloc2(localPerturbationU,localn,NoMembers,'localPerturbationU')
                call alloc2(localPerturbationV,localn,NoMembers,'localPerturbationV')
                call alloc2(localPerturbationPhi,localn,NoMembers,'localPerturbationPhi')

                ! \bar{x}_f^\ell: N_{state}^\ell
                call alloc1(localMeanU,localn,'localMeanU')
                call alloc1(localMeanV,localn,'localMeanV')
                call alloc1(localMeanPhi,localn,'localMeanPhi')

                ! Y_p^\ell: N_{obs}^\ell \times N_{ens}
                call alloc2(localEnsObsU,localn,NoMembers,'localEnsObsU')
                call alloc2(localEnsObsV,localn,NoMembers,'localEnsObsV')
                call alloc2(localEnsObsPhi,localn,NoMembers,'localEnsObsPhi')

                ! \bar{y}_f^\ell: N_{obs}^\ell
                call alloc1(localMeanObsU,localn,'localMeanObsU')
                call alloc1(localMeanObsV,localn,'localMeanObsV')
                call alloc1(localMeanObsPhi,localn,'localMeanObsPhi')

                ! C: N_{ens} \times N_{obs}^\ell
                call alloc2(localTempEnsObsU,NoMembers,localn,'localTempEnsObsU')
                call alloc2(localTempEnsObsV,NoMembers,localn,'localTempEnsObsV')
                call alloc2(localTempEnsObsPhi,NoMembers,localn,'localTempEnsObsPhi')

                ! X_a^\ell: N_{state}^\ell \times N_{ens}
                call alloc2(localAnalysisU,localn,NoMembers,'localAnalysisU')
                call alloc2(localAnalysisV,localn,NoMembers,'localAnalysisV')
                call alloc2(localAnalysisPhi,localn,NoMembers,'localAnalysisPhi')

                localn = 0

                do locali = max(2,i-NXCorrLength),min(n-1,i+NXCorrLength)
                    do localj = max(2,j-NYCorrLength),min(m-1,j+NYCorrLength)

                        dist=sqrt(abs((DX*(locali-i))**2+(DY*(localj-j))**2))

                        if (dist .le. CovLocalizationLength) then
                            localn = localn + 1

                            stateindex = map_2to1(locali-1,localj-1,n-2)

                            localPerturbationU(localn,:) = perturbationU(stateindex,:)
                            localPerturbationV(localn,:) = perturbationV(stateindex,:)
                            localPerturbationPhi(localn,:) = perturbationPhi(stateindex,:)

                            localMeanU(localn) = meanU(stateindex)
                            localMeanV(localn) = meanV(stateindex)
                            localMeanPhi(localn) = meanPhi(stateindex)

                            localEnsObsU(localn,:) = ensObsU(stateindex,:)
                            localEnsObsV(localn,:) = ensObsV(stateindex,:)
                            localEnsObsPhi(localn,:) = ensObsPhi(stateindex,:)

                            localMeanObsU(localn) = uobs(locali,localj,k)
                            localMeanObsV(localn) = vobs(locali,localj,k)
                            localMeanObsPhi(localn) = phiobs(locali,localj,k)
                        else
                            print *,'dist > CovLoc',dist,CovLocalizationLength
                        endif
                    enddo
                enddo

                localAnalysisCovarU = 0.
                localAnalysisCovarV = 0.
                localAnalysisCovarPhi = 0.

                ! \tilde{P}_a = \left[ \frac{N_{ens}}{\lambda}I + \left(Y_p^\ell\right)^\trans R^{-1} Y_p^\ell\right]^{-1}
                do member=1,NoMembers
                    localAnalysisCovarU(member,member) = 1./CovInflation
                    localAnalysisCovarV(member,member) = 1./CovInflation
                    localAnalysisCovarPhi(member,member) = 1./CovInflation
                enddo

                localAnalysisCovarU = localAnalysisCovarU + &
                    matmul(transpose(localEnsObsU)/(UObsStddev**2),localEnsObsU)

                localAnalysisCovarV = localAnalysisCovarV + &
                    matmul(transpose(localEnsObsV)/(VObsStddev**2),localEnsObsV)

                localAnalysisCovarPhi = localAnalysisCovarPhi + &
                    matmul(transpose(localEnsObsPhi)/(PhiObsStddev**2),localEnsObsPhi)

                call eigdecomp(NoMembers,localAnalysisCovarU,localEigvU)
                call eigdecomp(NoMembers,localAnalysisCovarV,localEigvV)
                call eigdecomp(NoMembers,localAnalysisCovarPhi,localEigvPhi)

                print *,'U Eigvalues:',minval(localEigvU),maxval(localEigvU)
                print *,'V Eigvalues:',minval(localEigvV),maxval(localEigvV)
                print *,'Phi Eigvalues:',minval(localEigvPhi),maxval(localEigvPhi)

                ! inverse of \tilde{P}_a^\ell is just the inverse eigenvalues
                do member=1,NoMembers
                    localEigvU(member) = max(1./localEigvU(member),MIN_EIGV)
                    localEigvV(member) = max(1./localEigvV(member),MIN_EIGV)
                    localEigvPhi(member) = max(1./localEigvPhi(member),MIN_EIGV)
                enddo

                localTempEnsEnsU = 0.
                localTempEnsEnsV = 0.
                localTempEnsEnsPhi = 0.

                ! get diagonal terms for W_a = sqrt((NoMembers-1)*localEigv) and \tilde{P}_a
                do member=1,NoMembers
                    localTempEnsEnsU(member,member) = sqrt(localEigvU(member))
                    localTempEnsEnsV(member,member) = sqrt(localEigvV(member))
                    localTempEnsEnsPhi(member,member) = sqrt(localEigvPhi(member))

                    localTempEnsEns2U(member,member) = localEigvU(member)
                    localTempEnsEns2V(member,member) = localEigvV(member)
                    localTempEnsEns2Phi(member,member) = localEigvPhi(member)
                enddo

                ! complete W_a = LD^{1/2}L^T
                localTempEnsEnsU = matmul(localAnalysisCovarU, &
                    matmul(localTempEnsEnsU,transpose(localAnalysisCovarU)))

                localTempEnsEnsV = matmul(localAnalysisCovarV, &
                    matmul(localTempEnsEnsV,transpose(localAnalysisCovarV)))

                localTempEnsEnsPhi = matmul(localAnalysisCovarPhi, &
                    matmul(localTempEnsEnsPhi,transpose(localAnalysisCovarPhi)))

                ! fill out \tilde{P}_a: multiply by LL^T
                localAnalysisCovarU = matmul(localAnalysisCovarU, &
                    matmul(localTempEnsEns2U,transpose(localAnalysisCovarU)))

                localAnalysisCovarV = matmul(localAnalysisCovarV, &
                    matmul(localTempEnsEns2V,transpose(localAnalysisCovarV)))

                localAnalysisCovarPhi = matmul(localAnalysisCovarPhi, &
                    matmul(localTempEnsEns2Phi,transpose(localAnalysisCovarPhi)))

                ksiv = 0.
                minvar = 1
                print *,'Now minimizing U'
!                print*,'Now conducting alpha test of gradient correctness'
!                call alphaTest(100000.0d0)

                ! call MLEF minimization
                if (OptimizerNum .eq. 1) then ! L-BFGS

                    INCLUDE 'opt-lbfgs-mlef.incl'

                elseif (OptimizerNum .eq. 2) then ! LMBM

                    INCLUDE 'opt-lmbm-mlef.incl'

                elseif (OptimizerNum .eq. 3) then ! DESCON

                    INCLUDE 'opt-descon-mlef.incl'

                elseif (OptimizerNum .eq. 4) then ! CG-Descent

                    INCLUDE 'opt-cg_descent-mlef.incl'

                endif

                ksi(1:NoMembers) = ksiv(1:NoMembers)

                ksiv = 0.
                minvar = 2
                print *,'Now minimizing V'
!                print*,'Now conducting alpha test of gradient correctness'
!                call alphaTest(1000.0d0)

                ! call MLEF minimization
                if (OptimizerNum .eq. 1) then ! L-BFGS

                    INCLUDE 'opt-lbfgs-mlef.incl'

                elseif (OptimizerNum .eq. 2) then ! LMBM

                    INCLUDE 'opt-lmbm-mlef.incl'

                elseif (OptimizerNum .eq. 3) then ! DESCON

                    INCLUDE 'opt-descon-mlef.incl'

                elseif (OptimizerNum .eq. 4) then ! CG-Descent

                    INCLUDE 'opt-cg_descent-mlef.incl'

                endif

                ksi(NoMembers+1:2*NoMembers) = ksiv(1:NoMembers)

                ksiv = 0.
                minvar = 3
                print *,'Now minimizing Phi'
!                print*,'Now conducting alpha test of gradient correctness'
!                call alphaTest(1000.0d0)

                ! call MLEF minimization
                if (OptimizerNum .eq. 1) then ! L-BFGS

                    INCLUDE 'opt-lbfgs-mlef.incl'

                elseif (OptimizerNum .eq. 2) then ! LMBM

                    INCLUDE 'opt-lmbm-mlef.incl'

                elseif (OptimizerNum .eq. 3) then ! DESCON

                    INCLUDE 'opt-descon-mlef.incl'

                elseif (OptimizerNum .eq. 4) then ! CG-Descent

                    INCLUDE 'opt-cg_descent-mlef.incl'

                endif

                ksi(2*NoMembers+1:3*NoMembers) = ksiv(1:NoMembers)

                ! update the analysis state
                localMeanU = matmul(localPerturbationU/sqrt(NoMembers-1.), &
                             matmul(localTempEnsEnsU,ksi(1:NoMembers))) +  &
                             localMeanU

                localMeanV = matmul(localPerturbationV/sqrt(NoMembers-1.), &
                             matmul(localTempEnsEnsV,ksi(NoMembers+1:2*NoMembers))) + &
                             localMeanV

                localMeanPhi = matmul(localPerturbationPhi/sqrt(NoMembers-1.), &
                               matmul(localTempEnsEnsPhi,ksi(2*NoMembers+1:3*NoMembers))) + &
                               localMeanPhi

                ! redo W_a matrix with updated
                localAnalysisCovarU = 0.
                localAnalysisCovarV = 0.
                localAnalysisCovarPhi = 0.

                ! \tilde{P}_a = \left[ \frac{N_{ens}}{\lambda}I + \left(Y_p^\ell\right)^\trans R^{-1} Y_p^\ell\right]^{-1}
                do member=1,NoMembers
                    localAnalysisCovarU(member,member) = 1./CovInflation
                    localAnalysisCovarV(member,member) = 1./CovInflation
                    localAnalysisCovarPhi(member,member) = 1./CovInflation
                enddo

                ! update the mean obs
                do ln=1,localn

                    call obs_operator(localMeanU(ln), &
                                      localMeanV(ln), &
                                      localMeanPhi(ln), &
                                      uval,vval,phival)

                    localMeanObsU(ln) = uval
                    localMeanObsV(ln) = vval
                    localMeanObsPhi(ln) = phival

                enddo

                do ln=1,localn
                    do member=1,NoMembers

                        call obs_operator(localMeanU(ln) + localPerturbationU(ln,member)/sqrt(NoMembers-1.), &
                                          localMeanV(ln) + localPerturbationV(ln,member)/sqrt(NoMembers-1.), &
                                          localMeanPhi(ln) + localPerturbationPhi(ln,member)/sqrt(NoMembers-1.), &
                                          uval,vval,phival)

                        localEnsObsU(ln,member) = uval - localMeanObsU(ln)
                        localEnsObsV(ln,member) = vval - localMeanObsV(ln)
                        localEnsObsPhi(ln,member) = phival - localMeanObsPhi(ln)

                    enddo
                enddo

                localAnalysisCovarU = localAnalysisCovarU + &
                    matmul(transpose(localEnsObsU)/(UObsStddev**2),localEnsObsU)

                localAnalysisCovarV = localAnalysisCovarV + &
                    matmul(transpose(localEnsObsV)/(VObsStddev**2),localEnsObsV)

                localAnalysisCovarPhi = localAnalysisCovarPhi + &
                    matmul(transpose(localEnsObsPhi)/(PhiObsStddev**2),localEnsObsPhi)

                call eigdecomp(NoMembers,localAnalysisCovarU,localEigvU)
                call eigdecomp(NoMembers,localAnalysisCovarV,localEigvV)
                call eigdecomp(NoMembers,localAnalysisCovarPhi,localEigvPhi)

                ! inverse of \tilde{P}_a^\ell is just the inverse eigenvalues
!                do member=1,NoMembers
!                    localEigvU(member) = max(1./localEigvU(member),MIN_EIGV)
!                    localEigvV(member) = max(1./localEigvV(member),MIN_EIGV)
!                    localEigvPhi(member) = max(1./localEigvPhi(member),MIN_EIGV)
!                enddo
                ! inverse of \tilde{P}_a^\ell is just the inverse eigenvalues
                do member=1,NoMembers
                    localEigvU(member) = 1./max(localEigvU(member),MIN_EIGV)
                    localEigvV(member) = 1./max(localEigvV(member),MIN_EIGV)
                    localEigvPhi(member) = 1./max(localEigvPhi(member),MIN_EIGV)
                enddo

                localTempEnsEnsU = 0.
                localTempEnsEnsV = 0.
                localTempEnsEnsPhi = 0.

                ! get diagonal terms for W_a = sqrt(localEigv)
                do member=1,NoMembers
                    localTempEnsEnsU(member,member) = sqrt(localEigvU(member))
                    localTempEnsEnsV(member,member) = sqrt(localEigvV(member))
                    localTempEnsEnsPhi(member,member) = sqrt(localEigvPhi(member))
                enddo

                ! complete W_a = LD^{1/2}L^T
                localTempEnsEnsU = matmul(localAnalysisCovarU, &
                    matmul(localTempEnsEnsU,transpose(localAnalysisCovarU)))

                localTempEnsEnsV = matmul(localAnalysisCovarV, &
                    matmul(localTempEnsEnsV,transpose(localAnalysisCovarV)))

                localTempEnsEnsPhi = matmul(localAnalysisCovarPhi, &
                    matmul(localTempEnsEnsPhi,transpose(localAnalysisCovarPhi)))

                ! update the perturbation
                localPerturbationU = matmul(localPerturbationU,&
                                     transpose(localTempEnsEnsU))

                localPerturbationV = matmul(localPerturbationV,&
                                     transpose(localTempEnsEnsV))

                localPerturbationPhi = matmul(localPerturbationPhi,&
                                       transpose(localTempEnsEnsPhi))

                if (NXCorrLength .ge. n .and. NYCorrLength .ge. m) then
                    ln = 0

                    do locali = 2,n-1
                        do localj = 2,m-1
                            ln = ln + 1

                            EnsembleU(:, locali, localj, k) = &
                                localPerturbationU(ln,:) + &
                                localMeanU(ln)

                            EnsembleV(:, locali, localj, k) = &
                                localPerturbationV(ln,:) + &
                                localMeanV(ln)

                            EnsemblePhi(:, locali, localj, k) = &
                                localPerturbationPhi(ln,:) + &
                                localMeanPhi(ln)

                            analysisDone = .true.
                        enddo
                    enddo
                else
                    EnsembleU(:, i, j, k) = &
                        localPerturbationU(analysisPoint,:) + &
                        localMeanU(analysisPoint)

                    EnsembleV(:, i, j, k) = &
                        localPerturbationV(analysisPoint,:) + &
                        localMeanV(analysisPoint)

                    EnsemblePhi(:, i, j, k) = &
                        localPerturbationPhi(analysisPoint,:) + &
                        localMeanPhi(analysisPoint)
                endif

                deallocate(localPerturbationU)
                deallocate(localPerturbationV)
                deallocate(localPerturbationPhi)

                deallocate(localMeanU)
                deallocate(localMeanV)
                deallocate(localMeanPhi)

                deallocate(localEnsObsU)
                deallocate(localEnsObsV)
                deallocate(localEnsObsPhi)

                deallocate(localTempEnsObsU)
                deallocate(localTempEnsObsV)
                deallocate(localTempEnsObsPhi)

                deallocate(localMeanObsU)
                deallocate(localMeanObsV)
                deallocate(localMeanObsPhi)

                deallocate(localAnalysisU)
                deallocate(localAnalysisV)
                deallocate(localAnalysisPhi)

                ! end loop over grid points
            enddo
        enddo

        meanU = 0.
        meanV = 0.
        meanPhi = 0.

        do member=1,NoMembers
             call set_boundary_conditions(k, m, n, EnsembleU(  member, :, :, k), &
                                                   EnsembleV(  member, :, :, k), &
                                                   EnsemblePhi(member, :, :, k))

            do i=2,n-1
                do j=2,m-1
                    stateindex = map_2to1(i-1,j-1,n-2)

                    meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)
                enddo
            enddo
        enddo

        meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        do i=2,n-1
            do j=2,m-1
                stateindex = map_2to1(i-1,j-1,n-2)

                u(i,j,k) = meanU(stateindex)
                v(i,j,k) = meanV(stateindex)
                phi(i,j,k) = meanPhi(stateindex)

               call set_boundary_conditions(k, m, n, u(:,:,k), v(:,:,k), phi(:,:,k))

            enddo
        enddo

        call calculateRmse(k)

        write (86,*) k, sqrt(urmse)/rmseroot, sqrt(vrmse)/rmseroot, sqrt(phirmse)/rmseroot

        print *, 'Completed iteration', k
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

    print *, 'Final RMSE is: '

    do k=1,nt
        call calculateRmse(k)
    enddo

    close(86)

    call finalize()

    deallocate(perturbationU)
    deallocate(perturbationV)
    deallocate(perturbationPhi)

    deallocate(ensObsU)
    deallocate(ensObsV)
    deallocate(ensObsPhi)

    deallocate(meanU)
    deallocate(meanV)
    deallocate(meanPhi)

    deallocate(meanObsU)
    deallocate(meanObsV)
    deallocate(meanObsPhi)

    deallocate(localAnalysisCovarU)
    deallocate(localAnalysisCovarV)
    deallocate(localAnalysisCovarPhi)

    deallocate(localTempEnsEnsU)
    deallocate(localTempEnsEnsV)
    deallocate(localTempEnsEnsPhi)

    deallocate(localTempEnsEns2U)
    deallocate(localTempEnsEns2V)
    deallocate(localTempEnsEns2Phi)

    deallocate(localMeanTempEnsEnsU)
    deallocate(localMeanTempEnsEnsV)
    deallocate(localMeanTempEnsEnsPhi)

    deallocate(dtemp)
    deallocate(xtemp)
    deallocate(gtemp)
end program

subroutine eigdecomp(N,A,eigv)
    implicit none

    integer, intent(in) :: N
    double precision, intent(inout), dimension(N,N) :: A
    double precision, intent(out), dimension(N) :: eigv

    double precision, dimension(N,N) :: Atmp
    double precision, dimension(N) :: eigvtmp

    integer :: lwork, ierr
    double precision, dimension(1) :: temp
    double precision, dimension(:), allocatable :: work

    integer :: i,j

    Atmp = A

    ! query the optimal work space
    lwork = -1
    call DSYEV('V','U',N,Atmp,N,eigvtmp,temp,lwork,ierr)

    lwork = temp(1)

    allocate(work(lwork),stat=ierr)
    if (ierr.ne.0) then
        write(*,*) "Allocation error: work"
        STOP
    end if

    ! solve the eigenvalue problem
    call DSYEV('V','U',N,Atmp,N,eigvtmp,work,lwork,ierr)
    if (ierr.ne.0) then
        write(*,*) "Error solving the eigenvalue decomposition: ",ierr
        STOP
    end if

    do j=1,N
        eigv(j) = eigvtmp(N - j + 1)

        do i=1,N
            A(i,j) = Atmp(i,N - j + 1)
        enddo
    enddo

    deallocate(work)

end subroutine
